import argparse
import json
import os
import sys

import yaml

from xh_dict_utils.ExtractFromFile import SupportedFormat
from xh_dict_utils.Layer1AppEngine import AppEngine
from xh_dict_utils.command.ToArray import *
from xh_dict_utils.command.CommandTemplate import *
from xh_dict_utils.dict_utils import Entries, Selector


def main():
    PROGRAM_NAME = "yaml-json-modification"
    DESCRIPTION = "A simple program modify yaml or json file"
    appEngine = AppEngine(
        program_name=PROGRAM_NAME,
        description=DESCRIPTION,
    )

    class LoadPipe(CommandTemplate):
        def __init__(self):
            super().__init__("load-pipe", "load from pipe")

        def parserOp(self, parser: argparse.ArgumentParser):
            parser.add_argument("--format", choices=["json", "yaml"], default="yaml")

        def handleOp(self, argv: [str]):
            format_type = getattr(SupportedFormat, argv.format.upper())
            if format_type == SupportedFormat.YAML:
                d = yaml.safe_load(sys.stdin.read())
            elif format_type == SupportedFormat.JSON:
                d = json.loads(sys.stdin.read())
            else:
                raise Exception("Not supported format")
            yaml.dump(d, sys.stdout)

    appEngine.regSubCommandWithTemplate(LoadPipe())

    class LoadFile(CommandTemplate):
        def __init__(self):
            super().__init__("load-file", "load file")

        def parserOp(self, parser: argparse.ArgumentParser):
            parser.add_argument("--format", choices=["json", "yaml"], default="yaml")
            parser.add_argument("--file")

        def handleOp(self, argv: [str]):
            format = getattr(SupportedFormat, argv.format.upper())

            def read_all_in_file(filePath) -> str:
                if not os.path.exists(filePath):
                    raise Exception(f"Path[{filePath}] not exists")

                if not os.path.isfile(filePath):
                    raise Exception(f"Path[{filePath}] is not file")
                with open(filePath, "r") as f:
                    return f.read()

            data = read_all_in_file(argv.file)
            d = dict() if data == "" else json.loads(data)

            def printDict(d: dict):
                if format == SupportedFormat.YAML:
                    return yaml.dump(d, sys.stdout)
                elif format == SupportedFormat.JSON:
                    return json.dump(d, sys.stdout)
                else:
                    raise Exception("Not supported format")

            printDict(d)

    appEngine.regSubCommandWithTemplate(LoadFile())

    class UpsertNode(CommandTemplate):
        def __init__(self):
            super().__init__("upsert-node", "update or insert to a node")

        def parserOp(self, parser: argparse.ArgumentParser):
            parser.add_argument("--selector", type=str)
            parser.add_argument("--dict-from-file", type=str)
            parser.add_argument("--dict-from-pipe", action="store_true", default=False)

            parser.add_argument("--key", type=str)

            parser.add_argument("--node-value", type=str)
            parser.add_argument("--node-value-file", type=str)

            parser.add_argument("--node-value-type", type=str, default="str",
                                choices=["str", "int", "number", "object"])
            parser.add_argument("--node-value-type-str", dest="node_value_type", action="store_const", const="str")
            parser.add_argument("--node-value-type-int", dest="node_value_type", action="store_const", const="int")
            parser.add_argument("--node-value-type-number", dest="node_value_type", action="store_const",
                                const="number")
            parser.add_argument("--node-value-type-object", dest="node_value_type", action="store_const",
                                const="object")

        def handleOp(self, argv: [str]):
            selector = argv.selector

            dict_from_pipe = argv.dict_from_pipe
            dict_from_file = argv.dict_from_file
            node_value_file = argv.node_value_file
            node_value = argv.node_value
            node_value_type = argv.node_value_type
            key = argv.key

            node_value_from_pipe = True if node_value is None and node_value_file is None else False

            def validate():
                if node_value_from_pipe and dict_from_pipe:
                    raise Exception("Cannot read `dictionary` and `node-value` both from pipe")
                if not node_value_from_pipe:
                    node_value_file = argv.node_value_file
                    node_value = argv.node_value
                    if node_value_file is not None and node_value is not None:
                        raise Exception(
                            f"node-value-file[{node_value_file}] and node-value[{node_value}] is mutually exclusive")
                    if node_value_file is not None:
                        if not os.path.exists(node_value_file):
                            raise Exception(f"[{node_value_file}] not exists")
                        if not os.path.isfile(node_value_file):
                            raise Exception(f"[{node_value_file}] not a file")
                if not dict_from_pipe:
                    if not os.path.exists(dict_from_file):
                        raise Exception(f"[{dict_from_file}] not exists")
                    if not os.path.isfile(dict_from_file):
                        raise Exception(f"[{dict_from_file}] not a file")

            validate()

            d = yaml.safe_load(sys.stdin.read() if dict_from_pipe else open(dict_from_file).read())
            entries = Entries.from_dict(d).match_notation(selector)

            def load():
                if node_value is not None:
                    data = node_value
                elif node_value_file:
                    data = open(node_value_file, "r").read()
                elif node_value_from_pipe:
                    data = sys.stdin.read()
                else:
                    raise Exception("Not supported node-value type")

                if node_value_type == "str":
                    return data
                elif node_value_type == "int":
                    return int(data)
                elif node_value_type == "number":
                    return float(data)
                elif node_value_type == "object":
                    return yaml.safe_load(data)
                else:
                    raise Exception("Not supported node-value type")

            node_value_content = load()
            if entries.unique_result():
                head = entries.head()
                if type(head.value) is dict or type(head.value) is list:
                    if key is None:
                        raise Exception("key is empty")
                    head.value.update({key: node_value_content})
                else:
                    key, _ = head.selector.key_and_parent()
                    head.parent.value.update({key: node_value_content})
            elif entries.no_result():
                raise Exception("no record found")
            else:
                raise Exception("multiple record result")

            yaml.dump(d, sys.stdout)

    appEngine.regSubCommandWithTemplate(UpsertNode())

    class RemoveNode(CommandTemplate):
        def __init__(self):
            super().__init__("remove-node", "remove a node")

        def parserOp(self, parser: argparse.ArgumentParser):
            parser.add_argument("--selector", type=str)

        def handleOp(self, argv: [str]):
            selector = argv.selector
            if Selector.from_notation(selector).is_array_group():
                is_array_group = True
            else:
                is_array_group = False

            d = yaml.safe_load(sys.stdin.read())

            if is_array_group:
                entries = Entries.from_dict(d).match_exact(selector)
            else:
                entries = Entries.from_dict(d).match_notation(selector)

            for entry in entries.asGenerator():
                if entry.is_root():
                    raise Exception("Can not remove root element")

            for entry in entries.asGenerator():
                parent = entry.parent
                if type(parent.value) is dict:
                    key, _ = entry.selector.key_and_parent()
                    del parent.value[key]
                elif type(parent.value) is list and entry.selector.is_array_item():
                    index = entry.selector.get_array_index()
                    del parent.value[index]
                else:
                    raise Exception("Not support")
            print(d)

    appEngine.regSubCommandWithTemplate(RemoveNode())

    class Query(CommandTemplate):
        def __init__(self):
            super().__init__("query", "query by selector")

        def parserOp(self, parser: argparse.ArgumentParser):
            parser.add_argument("--selector", type=str)
            parser.add_argument("--output", type=str, default="{selector} -> {value}")
            parser.add_argument("--exact", dest="notation_type", action="store_const", const="exact")
            parser.add_argument("--regex", dest="notation_type", action="store_const", const="regex")
            parser.set_defaults(notation_type="notation")
            parser.set_defaults(exact=False)

        def handleOp(self, argv: [str]):
            selector = argv.selector
            if argv.notation_type == "exact":
                findings = Entries.from_dict(yaml.safe_load(sys.stdin)).match_exact(selector)
            elif argv.notation_type == "regex":
                findings = Entries.from_dict(yaml.safe_load(sys.stdin)).match_regex(selector)
            if argv.notation_type == "notation":
                findings = Entries.from_dict(yaml.safe_load(sys.stdin)).match_notation(selector)

            for finding in findings.asGenerator():
                print(argv.output.format_map(
                    {"selector": finding.selector.selector_str, "value": finding.value, "context": finding}))

    appEngine.regSubCommandWithTemplate(Query())

    appEngine.regSubCommandWithTemplate(ToArray())

    class Output(CommandTemplate):
        def __init__(self):
            super().__init__("output", "output as json")

        def parserOp(self, parser: argparse.ArgumentParser):
            parser.add_argument("--store", type=str, default=None)
            parser.add_argument("--format", choices=["json", "yaml"])
            parser.add_argument("--pretty", action="store_true", default=False)

        def handleOp(self, argv: [str]):
            class Logger(object):
                def __init__(self, channel: list):
                    self.channels = channel

                def write(self, message):
                    for channel in self.channels:
                        channel.write(message)

                def flush(self):
                    # this flush method is needed for python 3 compatibility.
                    # this handles the flush command by doing nothing.
                    # you might want to specify some extra behavior here.
                    pass

            if argv.store is None:
                sys.stdout = Logger([sys.stdout])
            else:
                sys.stdout = Logger([sys.stdout, open(argv.store, "w")])

            format_type = getattr(SupportedFormat, argv.format.upper())
            d = yaml.safe_load(sys.stdin)
            if d is None:
                d = {}
            if format_type == SupportedFormat.YAML:
                yaml.dump(d, sys.stdout)
            elif format_type == SupportedFormat.JSON:
                if argv.pretty:
                    json.dump(d, sys.stdout, indent=2)
                else:
                    json.dump(d, sys.stdout)
            else:
                raise Exception("Not supported format")

    appEngine.regSubCommandWithTemplate(Output())

    appEngine.process(sys.argv)



import argparse
from abc import ABC, abstractmethod


class CommandTemplate(ABC):
    def __init__(self, name: str, help: str):
        self.name = name
        self.help = help

    @abstractmethod
    def parserOp(self, parser: argparse.ArgumentParser):
        pass

    @abstractmethod
    def handleOp(self, argv: [str]):
        pass




from enum import Enum
from pathlib import Path


class SupportedFormat(Enum):
    JSON = "json"
    YAML = "yaml"


def getKey(args):
    targetFile = Path(args.file)

    if not targetFile.exists():
        raise Exception(f'Path[{targetFile}] not exists')


def readFromFile(filePath):
    with open(filePath, "r") as f:
        while True:
            line = f.readline()
            if line:
                yield line
            else:
                break


def traverseNode(root, nodeNotation, do=lambda parent, nodeName: (parent, nodeName)):
    if root == None:
        raise Exception("Node input is empty")

    if nodeNotation and (not nodeNotation == ""):
        levels = nodeNotation.split(".")
        nodeName = levels[-1]

        parent = root
        for x in levels[:-1]:
            if not x in parent:
                parent[x] = dict()
            parent = parent[x]

        return True, do(parent, nodeName)
    else:
        return False, None, None





import re
from dataclasses import dataclass
from enum import Enum
from typing import Generator, Callable, Union, Tuple


class EntryType(Enum):
    DICT = 0
    LIST = 1
    LIST_ITEM = 2
    VALUE = 3


class Selector:
    __ROOT_SELECTOR: "Selector" = None
    selector_str: str

    def __init__(self, selector_str: str):
        self.selector_str = selector_str

    def is_root(self):
        return self.selector_str == ""

    def is_not_root(self):
        return not self.is_root()

    def is_array(self):
        compiled = re.compile("^.*\\.\\[(\\d+|\\*)]$")
        return compiled.match(self.selector_str) is not None

    def is_array_group(self):
        compiled = re.compile("^.*\\.\\[\\*]$")
        return compiled.match(self.selector_str) is not None

    def is_array_item(self):
        compiled = re.compile("^.*\\.\\[\\d+]$")
        return compiled.match(self.selector_str) is not None

    def get_array_index(self):
        if self.is_array_item():
            return int(self.selector_str.split(".")[-1][1:-1])
        else:
            raise Exception("Not an array item")

    def is_object_or_value(self):
        return not self.is_array()

    @staticmethod
    def root():
        if Selector.__ROOT_SELECTOR is None:
            Selector.__ROOT_SELECTOR = Selector("")
        return Selector.__ROOT_SELECTOR

    @staticmethod
    def from_notation(notation_str: str) -> 'Selector':
        return Selector(notation_str)

    def depth(self):
        return len(self.dot_chain())

    def object_or_value(self, key: str):
        return Selector(f"{key}") if self.is_root() else Selector(f"{self.selector_str}.{key}")

    def list_item(self, key: int):
        ends_with = ".[*]"
        offset = len(ends_with) * -1
        if self.selector_str.endswith(ends_with):
            return Selector(f"[{key}]") if self.is_root() else Selector(f"{self.selector_str[:offset]}.[{key}]")
        else:
            return Selector(f"[{key}]") if self.is_root() else Selector(f"{self.selector_str}.[{key}]")

    # def list_group(self):
    #     return Selector(f"[*]") if self.is_root() else Selector(f"{self.selector_str}.[*]")

    def list_group(self, index: int):
        return Selector(f"{index}.[*]") if self.is_root() else Selector(f"{self.selector_str}.{index}.[*]")

    def pop(self):
        return self.key_and_parent()[1]

    def dot_chain(self) -> [str]:
        return [i for i in self.selector_str.split(".") if i != "[*]"]
    def key_and_parent(self):
        splited = self.dot_chain()
        if self.is_root():
            return Selector.root(), None
        elif len(splited) == 1:
            return splited[0], Selector.root()
        else:
            return splited[-1], Selector(".".join(splited[:-1]))


@dataclass
class Entry:
    selector: Selector
    value: any
    parent: Union["Entry", None]

    def get_entry_type(self):
        if type(self.value) is dict:
            return EntryType.DICT
        elif type(self.value) is list:
            return EntryType.LIST
        else:
            return EntryType.VALUE

    def get_selector(self):
        return self.selector.selector_str

    def is_not_root(self):
        return self.selector.is_not_root()

    def is_root(self):
        return self.selector.is_root()

    def get_dict(self, key: str):
        return Entry(self.selector.object_or_value(key), {}, self)

    def get_list(self, key: int):
        return Entry(self.selector.list_item(key), [], self)

    def get_any_list(self):
        return Entry(self.selector.list_group(), [], self)

    def get_value(self, key: str, value: any):
        return Entry(self.selector.object_or_value(key), value, self)

    def match(self, matching: Callable[["Entry"], bool]) -> bool:
        return matching(self)

    def match_notation(self, selector: str):
        if selector == "":
            return self.is_root()
        notation_reg = selector.replace("[*]", "\\[[\\d|\\*]+\\]")
        return self.regex_selector(notation_reg)

    def exact_selector(self, selector: str):
        return self.match(lambda entry: entry.selector.selector_str == selector)

    def regex_selector(self, selector: str):
        compiled_selector = re.compile(selector)
        return self.match(lambda select: compiled_selector.match(select.get_selector()) is not None)

    def getIndex(self):
        return self.selector.get_array_index()

    def isList(self):
        return self.get_entry_type() == EntryType.LIST

    def isDict(self):
        return self.get_entry_type() == EntryType.DICT

    def isValue(self):
        return self.get_entry_type() == EntryType.VALUE

    def __str__(self):
        if self.is_root():
            return f"Root({self.value})"
        else:
            return f"Entry({self.get_selector()}, {self.value})"

    def __repr__(self):
        return self.__str__()


class Entries:
    entries: [Entry]

    def __init__(self, entries: [Entry]):
        self.entries = entries

    def match_exact(self, selector: str) -> "Entries":
        return Entries([t for t in self.entries if t.exact_selector(selector)])

    def match_regex(self, selector: str) -> "Entries":
        return Entries([t for t in self.entries if t.regex_selector(selector)])

    def match_notation(self, selector: str) -> "Entries":
        return Entries([t for t in self.entries if t.match_notation(selector)])

    def asGenerator(self) -> Generator[Entry, None, None]:
        yield from self.entries

    def asList(self) -> Generator[Entry, None, None]:
        yield from self.entries

    def size(self):
        return len(self.entries)

    def head_and_tail(self) -> Tuple[Entry, "Entries"]:
        head = self.entries[0] if len(self.entries) > 0 else None
        tail = Entries(self.entries[1:] if len(self.entries) > 1 else [])
        return head, tail

    def head(self):
        head, _ = self.head_and_tail()
        return head

    def tail(self):
        _, tail = self.head_and_tail()
        return tail

    def unique_result(self):
        return self.size() == 1

    def no_result(self):
        return self.size() == 0

    def multi_result(self):
        return self.size() > 1

    @staticmethod
    def from_dict(someNode: any) -> "Entries":
        if someNode is None:
            raise Exception(f"Input node is None")

        def innerFlatten(innerSomeNode: any, parent: Entry = None) \
                -> Generator[Entry, None, None]:
            if type(innerSomeNode) is dict:
                yield parent
                for key, value in innerSomeNode.items():
                    if type(value) is list:
                        entry = Entry(parent.selector.list_group(key), value, parent)
                        yield from innerFlatten(value, entry)
                    else:
                        entry = Entry(parent.selector.object_or_value(key), value, parent)
                        yield from innerFlatten(value, entry)
            elif type(innerSomeNode) is list:
                yield parent
                for index, value in enumerate(innerSomeNode):
                    entry = Entry(parent.selector.list_item(index), value, parent)
                    yield from innerFlatten(value, entry)
            else:
                yield parent

        return Entries(list(innerFlatten(someNode, Entry(Selector.root(), someNode, None))))

    @staticmethod
    def select_by_notation(dictionary: dict, notation: str) -> "Entries":
        return Entries.from_dict(dictionary).match_notation(notation)

    def __str__(self):
        return "\n".join([str(entry) for entry in self.entries])

    def __repr__(self):
        return self.__str__()



import json
import re
import sys
from typing import List

import yaml

from xh_dict_utils.command.CommandTemplate import *


class ToArray(CommandTemplate):
    def __init__(self):
        super().__init__("to-list", "row data to array")

    def parserOp(self, parser: argparse.ArgumentParser):
        parser.add_argument("--regex", type=str)
        parser.add_argument("--formatting", type=str, default="{{\"{key}\": \"{value}\"}}")

    def handleOp(self, argv: List[str]):
        regex_pattern = re.compile(argv.regex)
        formatting = argv.formatting
        pairs = ",".join([
            formatting.format_map(match_result.groupdict())
            for line in sys.stdin.readlines()
            for match_result in [regex_pattern.match(line)]
            if match_result
        ])
        yaml.dump(
            json.loads(f"[{pairs}]"), sys.stdout
        )


if __name__ == '__main__':
    toArray = ToArray()
    parser = argparse.ArgumentParser(
        prog="temp"
    )
    toArray.parserOp(parser)
    argv = parser.parse_args(["--regex", "(?P<key>[^ ]+)[ ]*=[ ]*\"(?P<value>.+)\"[ ]*"])
    toArray.handleOp(argv)
